#!python3

"""
This script reads the file "attributes.desc", which contains descriptions
of all multi-valued AST attributes, and creates a pair of files for each
attribute. For attribute "<attr>", the two files are:

<attr>_def.c: Contains the prototypes and definitions of methods to access
the attribute values.

<attr>_desc.c: Contains the description of the attribute getters and
setters that need to be included in th Python TypeObject for the class.
"""

from __future__ import print_function

import os
import os.path
import glob

def make_attributes( dirname=None ):

#  The hard-wired maximum number of dimensions supported by pyast.
   mxdim = 40

#  Open the input file.
   file = "attributes.desc"
   if dirname is not None:
      file = os.path.join( dirname, file )

   infile = open( file, "r" )

#  Read the input file.
   for line in infile:

#  Ignore blank lines or comment lines
      line = line.strip()
      if len(line) == 0 or line.startswith("#") :
         continue

#  Split the line into comma-separated fields
      (classname,attname,readonly,atype,desc,maxindex,minindex,items) = line.split(",")

#  Convert strings to numerical values
      minindex = int(minindex)

#  Initialise lists
      att_decs = list()
      att_descs = list()

#  Convert the fields to more useful types.
      items = items.split()
      if maxindex == "MXDIM":
         maxindex = mxdim
      else:
         maxindex = int(maxindex)

#  Loop over all indices for multi-valued attributes
      for i in range(minindex,maxindex+1):

#  Loop over all keys for multi-valued attributes
         for item in items:

#  Construct the name of the attribute.
            aname = attname;
            if item != "-":
               aname += "_" + item

            else:
               aname += "_" + str(i)

#  Form the C macro invocation that defines the attribute and append it
#  to the list.
            mac = "MAKE_GET"+readonly+atype+"("+classname+","+aname+")"
            att_decs.append(mac)

#  Form the attribute description to store in the Python TypeObject.
            mac = "DEFATT("+aname+", \""+desc+"\"),"
            att_descs.append(mac)

#  Multi-valued attributes can also (usually) be used without any index
#  of key. So add an unqualified attribute to the lists.
      mac = "MAKE_GET"+readonly+atype+"("+classname+","+attname+")"
      att_decs.append(mac)
      mac = "DEFATT("+attname+", \""+desc+"\"),"
      att_descs.append(mac)

#  Open the output def file
      cfilename = attname + "_def.c"
      if dirname is not None:
         cfilename = os.path.join( dirname, cfilename )
      cfile = open( cfilename, "w" )

# Write out a prologue for the C file
      print(r"""/*
*  Name:
*     """+ attname + "_def.c"+r"""

*  Purpose:
*     Declare and define the Python accessor methods for the """+attname+r"""
*     attribute.

*  Description:
*     This file uses the macros defined in pyast.h to define and declare
*     methods for accessing all the values within the multi-valued AST
*     attribute """+attname+r""".

*  Notes:
*     - This file is generated automatically by the "make_attributes.py"
*     script, and should not be edited. Instead, edit the file
*     attributes.desc which is used as input by "make_attributes.py".

*/
""", file=cfile)

#  Write out the attribute declarations, then close the file.
      for att in att_decs:
         print( att, file=cfile )
      cfile.close()

#  Open the output desc file
      cfilename = attname + "_desc.c"
      if dirname is not None:
         cfilename = os.path.join( dirname, cfilename )
      cfile = open( cfilename, "w" )

# Write out a prologue for the C file
      print(r"""/*
*  Name:
*     """+ attname + "_desc.c"+r"""

*  Purpose:
*     Add entries to the array of Python attribute accessors describing
*     the """+attname+r""" attribute.

*  Description:
*     This file uses the macros defined in pyast.h to create a description
*     of the multi-valued AST attribute """+attname+r""" for inclusion in
*     the array of attribute getters and setters store in the Python
*     TypeObject for the class.

*  Notes:
*     - This file is generated automatically by the "make_attributes.py"
*     script, and should not be edited. Instead, edit the file
*     attributes.desc which is used as input by "make_attributes.py".

*/
""", file=cfile)

#  Write out the attribute dedriptions, then close the file.
      for att in att_descs:
         print( "   "+att, file=cfile )
      cfile.close()

#  Close the input file.
   infile.close()


if __name__ == "__main__":
    make_attributes()



import starlink.Ast as Ast
import starlink.Grf as Grf
import matplotlib.pyplot as plt
try:
    import astropy.io.fits as pyfits
except ImportError:
    import pyfits

"""
This module provides function and classes that wrap up sequences of PyAST
calls to perform commonly used operations. It requires the PyFITS and
matplotlib libraries to be installed.
"""

# ======================================================================
class PyFITSAdapter:
   """

   Adapter to allow PyAST FitsChan objects to read and write headers to
   and from a PyFITS HDU.

   This class allows a PyFITS HDU to be used as the source or sink object
   with a FitsChan.

   When used as a FitsChan source, the PyFITSAdapter will allow the
   FitsChan to read each of the cards in the associated PyFITS header,
   thus allowing the cards to be copied into the FitsChan. This happens
   when the newly created, empty FitsChan is used for the first time
   (i.e. when any of its methods is invoked), and subsequently whenever
   the FitsChan.readfits() method is invoked.

   When used as a FitsChan sink, the PyFITSAdapter will allow the
   FitsChan to copy its own header cards into the PyFITS header. This
   happens when the FitsChan is deleted or when the FitsChan.writefits()
   method is invoked. If the PyFITSAdapter.Clear property is true, then
   the PyFITS header is first emptied of all existing headers, and the
   contents of the FitsChan are then stored in the PyFITS header.  If the
   PyFITSAdapter.Clear property is False, the original contents of the
   PyFITS header are retained. In this case each FitsChan card replaces any
   existing card that refers to the same keyword (if there is no card
   for the keyword already in the PyFITS header, the FitsChan card will
   be appended to the end of the header).

   """

   def __init__(self,hdu,clear=True):
      """
      Construct a PyFITSAdapter for a specified PyFITS HDU.

      Parameters:
         hdu: An element of the hdulist associated with a FITS file
            opened using pyfits.open(). If the entire hdulist is supplied,
            rather than an element of the hdulist, then the primary HDU
            (element zero) will be used.
         clear: This value is only used when using the PyFITSAdapter as
            a sink function with a FitsChan. If "clear" is True, the
	    original contents of the PyFITS header are deleted
	    immediately before the first FitsChan card is copied into the
	    PyFITS header. If "clear" is False, the original contents of
	    the PyFITS heasder are retained, with the FitsChan keywords
	    over-writing any existing values for the same keywords.

      Examples:
         - To read WCS from the 'DATA' extension in FITS file 'test.fit':

         >>> import pyfits
         >>> import starlink.Ast as Ast
         >>> import starlink.Atl as Atl
         >>>
         >>> hdulist = pyfits.open('test.fit')
         >>> fc = Ast.FitsChan( Atl.PyFITSAdapter( hdulist['DATA'] ) )
         >>> framset = fc.read()

         - To write a FrameSet to the primary HDU in FITS file 'old.fit',
         using standard FITS-WCS keywords, emptying the primtary HDU
         header first:

         >>> import pyfits
         >>> import starlink.Ast as Ast
         >>> import starlink.Atl as Atl
         >>>
         >>> hdulist = pyfits.open('old.fit')
         >>> fc = Ast.FitsChan( None, Atl.PyFITSAdapter( hdulist ) )
         >>> if fc.write( framset ) == 0:
         >>>    print("Failed to convert FrameSet to FITS header")
      """

#  If the supplied object behaves like a sequence, use element zero (the
#  primary HDU). Otherwise use the supplied object.
      try:
         self.hdu = hdu[ 0 ]
      except TypeError:
         self.hdu = hdu

#  Initialise the index of the next card to read or write.
      self.index = 0

#  Record whether the PyFITS header should be emptied before writing to
#  it for the first time
      self.clear = clear


# -----------------------------------------------------------------
   def astsource(self):

      """
      This method is called by the FitsChan to obtain a single 80-character
      FITS header card. It iterates over all the cards in the PyFITS
      header, returning each one in turn. It then returns "None" to
      indicate that there are no more header cards to read.
      """

      if self.index < len( self.hdu.header.ascard ):
         result = self.hdu.header.ascard[ self.index ].ascardimage()
         self.index += 1
      else:
         result = None
         self.index = 0

      return result

# -----------------------------------------------------------------
   def astsink(self,card):

      """
      This method is called by the FitsChan to store a single 80-character
      FITS header card. On the first invocation all cards will be deleted
      form the header if the "clear" property is true. Otherwise, if the
      header already contains a card for the keyword, the existing card
      is replaced with the new card. Otherwise, the new card is stored at
      the end of the header.
      """

      if self.index == 0 and self.clear:
         self.hdu.header = pyfits.Header()
         self.hdu.header.clear()

      card = pyfits.Card.fromstring(card)
      if card.key == "" or card.key == "BLANK":
         self.hdu.header.add_blank()
      elif card.key == "COMMENT":
         self.hdu.header.add_comment(card.value)
      elif card.key == "HISTORY":
         self.hdu.header.add_history(card.value)
      else:
         self.hdu.header.update( card.key, card.value, card.comment )

      self.index += 1


# ======================================================================
def readfitswcs( hdu ):

   r"""Reads an AST FrameSet from a FITS header.

      The header from the specified FITS HDU is read, and an AST FrameSet
      describing the WCS information in the header is returned. None is
      returned instead of a FrameSet if WCS information cannot be read
      from the header. A string identifying the scheme used to describe
      WCS information in the header (the encoding) is also returned.

      (frameset,encoding) = starlink.Atl.readfitswcs( hdu )

      Parameters:
         hdu: An element of the hdulist associated with a FITS file
 	    opened using pyfits.open(). If the entire hdulist is supplied,
 	    rather than an element of the hdulist, then the primary HDU
 	    (element zero) will be used.
         frameset: A reference to the FrameSet describing the pixel and
	    world coordinate systems read from the FITS header, or "None"
	    if no WCS could be read.
         encoding: Indicates how the WCS information was encoded in the
	    header. For possible values, see the documentation for the
	    "Encoding" attribute in SUN/211.

      Example:
         >>> import pyfits
         >>> import starlink.Atl as Atl
         >>>
         >>> hdulist = pyfits.open( 'test.fit' )
         >>> (frameset,encoding) = Atl.readfitswcs( hdulist[ 3 ] )
         >>> if frameset == None:
         >>>    print( "Cannot read WCS from test.fit" )

   """

   try:
      myhdu = hdu[ 0 ]
   except TypeError:
      myhdu = hdu

   fitschan = Ast.FitsChan( PyFITSAdapter( myhdu ) )
   encoding = fitschan.Encoding
   frameset = fitschan.read()
   return (frameset,encoding)




# ======================================================================
def writefitswcs( frameset, hdu, encoding="FITS-WCS" ):

   r"""Write an AST FrameSet to a FITS file.

      The WCS information described by the supplied FrameSet is converted
      into a set of FITS header cards which are stored in the supplied
      HDU (all cards in the header are first removed).

      nobj = starlink.Atl.writefitswcs( frameset, hdu, encoding="FITS-WCS" )

      Parameters:
         frameset: A reference to the FrameSet to be written out to the
            FITS header.
         hdu: An element of the PyFITS hdulist associated with a FITS file.
            The header cards generated from the FrameSet are stored in the
            header associated with this HDU. All cards are first removed
            from the header. If an entire hdulist is supplied, rather than
            an element of the hdulist, then the primary HDU (element zero)
            will be used.
         encoding: Indicates how the WCS information is to be encoded in the
	    header. For possible values, see the documentation for the
	    "Encoding" attribute in SUN/211.
         nobj:
            Returned equal to 1 if the FrameSet was converted successfully
            to FITS headers using the requested encoding, and zero
            otherwise.
      Example:
         >>> import starlink.Atl as Atl
         >>>
         >>> (frameset,encoding) = Atl.readfitswcs( hdu1 )
         >>> if Atl.writefitswcs( frameset, hdu2, encoding="FITS-AIPS" ) == 0:
         >>>    print( "Cannot convert WCS to FITS-AIPS encoding" )

    """

   fitschan = Ast.FitsChan( None, PyFITSAdapter(hdu) )
   fitschan.Encoding = encoding
   return fitschan.write( frameset )



# ======================================================================
def plotframeset( axes, gbox, bbox, frameset, options="" ):
   r"""Plot an annotated coordinate grid in a matplotlib axes area.

      plot = starlink.Atl.plotframeset( axes, gbox, bbox, frameset,
                                        options="" )

      Parameters:
         axes: A matplotlib "Axes" object. The annotated axes normally
            produced by matplotlib will be removed, and axes will
            instead be drawn by the AST Plot class.
         gbox: A list of four values giving the bounds of the new
            annotated axes within the matplotlib Axes object. The supplied
            values should be in the order (xleft,ybottom,xright,ytop) and
            should be given in the matplotlib "axes" coordinate system.
         bbox: A list of four values giving the bounds of the new
            annotated axes within the coordinate system represented by the
            base Frame of the supplied FrameSet. The supplied values should
            be in the order (xleft,ybottom,xright,ytop).
         frameset: An AST FrameSet such as returned by the Atl.readfitswcs
            function. Its base Frame should be 2-dimensional.
         options: An optional string holding a comma-separated list of Plot
            attribute settings. These control the appearance of the
            annotated axes.
         plot: A reference to the Ast.Plot that was used to draw the axes.

      Example:
         >>> import pyfits
         >>> import starlink.Atl as Atl
         >>> import matplotlib.pyplot
         >>>
         >>> hdulist = pyfits.open( 'test.fit' )
         >>> (frameset,encoding) = starlink.Atl.readfitswcs( hdulist[0] )
         >>> if frameset != None:
         >>>    naxis1 = hdulist[0].header['NAXIS1']
         >>>    naxis2 = hdulist[0].header['NAXIS2']
         >>>    Atl.plotframeset( matplotlib.pyplot.figure().add_subplot(111),
         >>>                      [ 0.1, 0.1, 0.9, 0.9 ],
         >>>                      [ 0.5, 0.5, naxis1+0.5, naxis2+0.5 ], frameset )
         >>>    matplotlib.pyplot.show()
   """

   axes.xaxis.set_visible( False )
   axes.yaxis.set_visible( False )

   plot = Ast.Plot( frameset, gbox, bbox, Grf.grf_matplotlib( axes ), options )
   plot.grid()
   return plot



# ======================================================================
def plotfitswcs( axes, gbox, hdu, options="" ):

   r"""Read WCS from a PyFITS HDU and plot an annotated coordinate grid
      in a matplotlib axes area. The grid covers the entire image.

      plot = starlink.Atl.plotfitswcs( axes, gbox, hdu, options="" )

      Parameters:
         axes: A matplotlib "Axes" object. The annotated axes normally
            produced by matplotlib will be removed, and axes will
            instead be drawn by the AST Plot class.
         gbox: A list of four values giving the bounds of the new
            annotated axes within the matplotlib Axes object. The supplied
            values should be in the order (xleft,ybottom,xright,ytop) and
            should be given in the matplotlib "axes" coordinate system.
         hdu: An element of the hdulist associated with a FITS file
 	    opened using pyfits.open(). If the entire hdulist is supplied,
 	    rather than an element of the hdulist, then the primary HDU
 	    (element zero) will be used.
         options: An optional string holding a comma-separated list
            of Ast.Plot attribute settings. These control the appearance
            of the annotated axes.
         plot: A reference to the Ast.Plot that was used to draw the axes.

      Example:
         >>> import pyfits
         >>> import starlink.Atl as Atl
         >>> import matplotlib.pyplot
         >>>
         >>> hdulist = pyfits.open( 'test.fit' )
         >>> Atl.plotfitswcs( matplotlib.pyplot.figure().add_subplot(111),
         >>>                  [ 0.1, 0.1, 0.9, 0.9 ], hdulist )
         >>> matplotlib.pyplot.show()
   """

   try:
      myhdu = hdu[ 0 ]
   except TypeError:
      myhdu = hdu

   (frameset,encoding) = readfitswcs( myhdu )
   naxis1 = myhdu.header[ 'NAXIS1' ]
   naxis2 = myhdu.header[ 'NAXIS2' ]
   return plotframeset( axes, gbox, [ 0.5, 0.5, naxis1+0.5, naxis2+0.5 ],
                        frameset, options )





import starlink.Ast as Ast
import matplotlib.pyplot
import matplotlib.lines
import math

"""
This module provides classes that provide primitive drawing facilities
for the starlink.Ast.Plot class. Each class defined in this module
implements the drawing methods required by the Plot class, using a
different underlying graphics system.

Currently only one class is provided, which uses the matplotlib
package to provide the primitive rdawing capabilities.

For further information about the methods implemented by these
classes, see the file grf_pgplot.c included in the AST source
distribution.
"""

class grf_matplotlib(object):

   """
   When creating a grf_matplotlib, the supplied "axes" object should be an
   instance of the matplotlib Axes class (or a subclass).
   """

#------------------------------------------------------------------------
   def __init__(self,axes):
      if isinstance(axes,matplotlib.axes.Axes):
         self.axes = axes

#  Save the current axis scales.
         self.Scales()

#  Create a temporary text string and line from which we can determine
#  the default graphics properties.
         xl,xr = self.axes.get_xlim()
         yb,yt = self.axes.get_ylim()
         xc = 0.5*(xl+xr);
         yc = 0.5*(yt+yb);
         text = matplotlib.text.Text( xc, yc, "a")
         line = matplotlib.lines.Line2D( [xc], [yc], marker="+")

#  Save the current default marker and text sizes.
         self.__deftsize = text.get_size()
         self.__defmsize = line.get_markersize()

#  Save the default text colour.
         defcol = text.get_color()

#  Save the default text font family and style.
         deffont = {"family":text.get_family(),"style":text.get_style()}

#  Save the default line style
         defstyle = line.get_linestyle()

#  A list used to convert AST integer marker types into matplotlib
#  character marker types.
         self.markers = ['s','.','+','*','o','x',',','^','v','<','>',
                         'p','h','D']

#  A list used to convert AST integer line style types into corresponding matplotlib
#  properties. Ensure the first line style is the default.
         self.styles = [ {"linestyle":defstyle}, {"linestyle":'-'},
                         {"linestyle":'--'}, {"linestyle":':'},
                         {"linestyle":'-.'} ]

#  A list used to convert AST integer font types into corresponding matplotlib
#  properties. Ensure the first font is the default.
         self.fonts = [ deffont, {"family":'serif',"style":'normal'},
                        {"family":'serif',"style":'italic'},
                        {"family":'sans-serif',"style":'normal'},
                        {"family":'sans-serif',"style":'italic'},
                        {"family":'monospace',"style":'normal'},
                        {"family":'monospace',"style":'italic'} ]

#  A list used to convert AST integer colours into corresponding matplotlib
#  properties. Ensure the first colour is the default.
         self.colours = [ {"color":defcol}, {"color":'red'}, {"color":'green'},
                          {"color":'blue'}, {"color":'cyan'},
                          {"color":'magenta'}, {"color":'yellow'},
                          {"color":'black'}, {"color":'darkgrey'},
                          {"color":'grey'}, {"color":'lightgrey'},
                          {"color":'white'} ]

#  The current graphics attribute values as used by AST
         self.__attrs = { Ast.grfLINE:{Ast.grfSTYLE:1, Ast.grfWIDTH:1, Ast.grfSIZE:1,
                                       Ast.grfFONT:1, Ast.grfCOLOUR:1},
                          Ast.grfMARK:{Ast.grfSTYLE:1, Ast.grfWIDTH:1, Ast.grfSIZE:1,
                                       Ast.grfFONT:1, Ast.grfCOLOUR:1},
                          Ast.grfTEXT:{Ast.grfSTYLE:1, Ast.grfWIDTH:1, Ast.grfSIZE:1,
                                       Ast.grfFONT:1, Ast.grfCOLOUR:1}}

#  The corresponding graphics properties used by matplotlib
         self.__props = { Ast.grfLINE:{}, Ast.grfMARK:{}, Ast.grfTEXT:{}}

#  Ensure the defaults are current.
         for attr in ( Ast.grfCOLOUR, Ast.grfWIDTH, Ast.grfSIZE,
                       Ast.grfFONT, Ast.grfSTYLE ):
            for prim in ( Ast.grfTEXT, Ast.grfLINE, Ast.grfMARK ):
               self.Attr( attr, 1.0, prim )

#  Set new delimiters for graphical sky axis values, using appropriate escape
#  sequences to get he superscripts looking nice.
         Ast.tunec( "hrdel", "%-%^85+%s70+h%>45+%+" )
         Ast.tunec( "mndel", "%-%^85+%s70+m%>45+%+" )
         Ast.tunec( "scdel", "%-%^85+%s70+s%>45+%+" )
         Ast.tunec( "dgdel", "%-%^90+%s60+o%>45+%+" )
         Ast.tunec( "amdel", "%-%^30+%s85+'%>45+%+" )
         Ast.tunec( "asdel", "%-%^30+%s85+\"%>45+%+" )
         Ast.tunec( "exdel", "10%-%^85+%s60+%>20+" )

#  Initialise the correction vector for text.
         self._xcorr = 0.0
         self._ycorr = 0.0

#  Save the current character heights, and update the vertical offset
#  correction for text.
         self.Qch()

#  Report an error if the supplied object is not suitable
      else:
         m = axes.__class__.__module__
         c = axes.__class__.__name__
         if m != "builtins":
            c = m + "." + c
         raise TypeError("The supplied axes object is a "+c+", it should "
                         "an instance of matplotlib.axes.Axes or a subclass")

#------------------------------------------------------------------------
   def Attr( self, attr, value, prim ):

#  Save the old AST attribute value.
      oldval = self.__attrs[prim][attr]

#  Nothing more to do if the new value is AST__BAD
      if value != Ast.BAD:

#  Save the old AST attribute value, and record the new value (if not .
         oldval = self.__attrs[prim][attr]
         self.__attrs[prim][attr] = value

#  Now need to update the matplotlib properties to make them reflect the
#  new AST value.

#  Style only applied to lines
         if attr == Ast.grfSTYLE:
            value = int(value) - 1
            if prim == Ast.grfLINE:
               if value >= 0 and value < len(self.styles):
                  self.__props[prim].update(self.styles[value])

#  Width applies to lines, marks and texts
         elif attr == Ast.grfWIDTH:
            if prim == Ast.grfLINE:

#  Get bounds of plot in user coords
               xl,xr = self.axes.get_xlim()
               yb,yt = self.axes.get_ylim()

#  Transform to device coords
               tr = self.axes.transData
               xl,yb = tr.transform([xl,yb])
               xr,yt = tr.transform([xr,yt])

#  Transform to inches
               tr = self.axes.get_figure().dpi_scale_trans.inverted()
               xl,yb = tr.transform([xl,yb])
               xr,yt = tr.transform([xr,yt])

#  Find the length of the diagonal in inches, and convert to points.
               diag = 72.0*math.sqrt( (xr-xl)**2 + (yt-yb)**2 )

#  Find the number of points corresponding to an AST line width of 1.0,
#  ensure it is at least 1 point, and then scale it by the supplied AST
#  line width.
               lw = max( 0.0005*diag, 1.0 )*value
               self.__props[prim].update({"linewidth":lw})

#  Similar for markers
            elif prim == Ast.grfMARK:
               xl,xr = self.axes.get_xlim()
               yb,yt = self.axes.get_ylim()
               tr = self.axes.transData
               xl,yb = tr.transform([xl,yb])
               xr,yt = tr.transform([xr,yt])
               tr = self.axes.get_figure().dpi_scale_trans.inverted()
               xl,yb = tr.transform([xl,yb])
               xr,yt = tr.transform([xr,yt])
               diag = 72.0*math.sqrt( (xr-xl)**2 + (yt-yb)**2 )
               lw = max( 0.0005*diag, 1.0 )*value
               self.__props[prim].update({"markeredgewidth":lw})

#  Cannot control exact line width of texts, use weight instead.
            elif prim == Ast.grfTEXT:
               if value < 0.0:
                  wgt = 0.0
               elif value < 1.0:
                  wgt = 250.0*value
               elif value < 10.0:
                  wgt = (650.0*value + 1600.0 )/9.0
               else:
                  wgt = 900
               self.__props[prim].update({"weight":wgt})

#  Size applies to marks and texts
         elif attr == Ast.grfSIZE:
            if prim == Ast.grfMARK:
               self.__props[prim].update({"markersize":self.__defmsize*value})
            elif prim == Ast.grfTEXT:
               self.__props[prim].update({"size":self.__deftsize*value})

#  Font only applies to texts
         elif attr == Ast.grfFONT:
            value = int(value) - 1
            if prim == Ast.grfTEXT:
               if value >= 0 and value < len(self.fonts):
                  self.__props[prim].update(self.fonts[value])

#  Colour applied to them all
         elif attr == Ast.grfCOLOUR:
            value = int(value) - 1
            if value >= 0 and value < len(self.colours):
               self.__props[prim].update(self.colours[value])


      return oldval

#------------------------------------------------------------------------
   def BBuf( self ):
      return

#------------------------------------------------------------------------
   def Cap( self, cap, value ):
      if cap == Ast.grfSCALES:
         return 1
      elif cap == Ast.grfMJUST:
         return 1;
      elif cap == Ast.grfESC:
         return 0;
      else:
         return 0

#------------------------------------------------------------------------
   def EBuf( self ):
      return

#------------------------------------------------------------------------
   def Flush( self ):
      return

#------------------------------------------------------------------------
   def Line( self, n, x, y ):
      self.axes.add_line( matplotlib.lines.Line2D( x, y,
                                                **self.__props[Ast.grfLINE] ) )

#------------------------------------------------------------------------
   def Mark( self, n, x, y, type ):
      if type < 0 or type >= len(self.markers):
         marker= '.'
      else:
         marker = self.markers[ type ]

      props = self.__props[Ast.grfMARK].copy()
      props["linestyle"] = 'None'
      props["marker"] = marker
      self.axes.add_line( matplotlib.lines.Line2D( x, y, **props ) )

#------------------------------------------------------------------------
   def Qch( self ):
      xl,xr = self.axes.get_xlim()
      yb,yt = self.axes.get_ylim()
      x = 0.5*(xl+xr)
      y = 0.5*(yt-yb)

      a = self.TxExt( "a", x, y, "CC", 0, 1 )
      self._chh = max(a[4:]) - min(a[4:])

      a = self.TxExt( "a", x, y, "CC", 1, 0 )
      self._chv = max(a[:3]) - min(a[:3])

      self._xcorr = self._chv*0.15
      self._ycorr = self._chh*0.15

      return (self._chv,self._chh)

#------------------------------------------------------------------------
   def Scales( self ):
      xleft,xright = self.axes.get_xlim()
      ybot,ytop = self.axes.get_ylim()
      a = self.axes.transData.transform([(xleft,ybot),(xright,ytop)])
      self.__alpha = ( a[1][0] - a[0][0] ) / ( xright - xleft )
      self.__beta = ( a[1][1] - a[0][1] ) / ( ytop - ybot )
      return (self.__alpha,self.__beta)

#------------------------------------------------------------------------
   def Text( self, text, x, y, just, upx, upy, boxprops={} ):
      if just[0] == "T":
         va = "top"
      elif just[0] == "B":
         va = "baseline"
      elif just[0] == "M":
         va = "bottom"
      else:
         va = "center"
      if just[1] == "L":
         ha = "left"
      elif just[1] == "R":
         ha = "right"
      else:
         ha = "center"
      rot = math.atan2( -upx, upy )*Ast.DR2D

#  matplotlib always seems to plot each text string a little higher than
#  requested, sp correct the reference position by a small amount
#  determined empirically to produce visually better text positioning.
      uplen = math.sqrt( upx**2 + upy**2 )
      if uplen > 0.0:
         x -= upx*self._xcorr/uplen
         y -= upy*self._ycorr/uplen

      props = self.__props[Ast.grfTEXT].copy()
      props["verticalalignment"] = va
      props["horizontalalignment"] = ha
      props["rotation"] = rot
      props["rotation_mode"] = "anchor"
      if len(boxprops) > 0:
         props["bbox"] = boxprops

      otext = matplotlib.text.Text( x, y, text, **props )
      self.axes.add_artist( otext )
      return otext

#------------------------------------------------------------------------
   def TxExt( self, text, x, y, just, upx, upy ):
      otext = self.Text( text, x, y, just, upx, upy, boxprops={"boxstyle":"square,pad=0.0"} )
      renderer = self.axes.get_figure().canvas.get_renderer()
      otext.draw(renderer)
      pix_verts = otext.get_bbox_patch().get_verts()
      wcs_verts = otext.get_transform().inverted().transform(pix_verts)
      otext.remove()

      return (wcs_verts[0][0], wcs_verts[1][0], wcs_verts[2][0], wcs_verts[3][0],
              wcs_verts[0][1], wcs_verts[1][1], wcs_verts[2][1], wcs_verts[3][1])




import unittest
import starlink.Ast
import copy
import numpy
import math
import copy
import filecmp
import string
import sys
import os.path
import os

#  A class that defines Channel source and sink functions that store text
#  in an internal list.
class TextStream():

   def __init__( self ):
      self.reset_sink()
      self.reset_source()

   def reset_sink( self ):
      self.text = []

   def reset_source( self ):
      self.index = 0;

   def astsink( self, text ):
      self.text.append(text)

   def astsource( self ):
      if self.index < len( self.text ):
         result = self.text[ self.index ]
         self.index += 1
      else:
         result = None
      return result

   def get( self ):
      return copy.deepcopy(self.text)


#  A dummy object used to test channel error reporting.
class DummyStream():
   pass

#  A class that provides primitive graphics operatons for a Plot. All it
#  does is record the values supplied to it by the Plot class.
class DummyGrf():
   def __init__( self ):
      self.Reset()

   def Reset( self ):
      self.linex = []
      self.liney = []
      self.nline = 0
      self.markx = []
      self.marky = []
      self.markt = []
      self.nmark = 0
      self.attr = []
      self.value = []
      self.prim = []
      self.textt = []
      self.textx = []
      self.texty = []
      self.textj = []
      self.ntext = 0

   def Attr( self, attr, value, prim ):
      self.attr += [attr]
      self.value += [value]
      self.prim += [prim]
      return 1

   def BBuf( self ):
      return 0

   def Cap( self, cap, value ):
      if cap == starlink.Ast.grfSCALES:
         return 1
      elif cap == starlink.Ast.grfMJUST:
         return 0;
      elif cap == starlink.Ast.grfESC:
         return 0;
      else:
         return 0

   def EBuf( self ):
      return 0

   def Flush( self ):
      return 0

   def Line( self, n, x, y ):
      self.linex.extend(x)
      self.liney.extend(y)
      self.nline += n

   def Mark( self, n, x, y, type ):
      self.markx.extend(x)
      self.marky.extend(y)
      self.markt += (type,)
      self.nmark += n

   def Qch( self ):
      return (1.0,1.0)

   def Scales( self ):
      return (1.0,1.0)

   def Text( self, text, x, y, just, upx, upy ):
      self.textt += [text]
      self.textx += [x]
      self.texty += [y]
      self.textj += [just]
      self.ntext += 1

   def TxExt( self, text, x, y, just, upx, upy ):
      return (x-0.1,x+0.1,x+0.1,x-0.1,y-0.1,y-0.1,y+0.1,y+0.1)


#  Tester
class TestAst(unittest.TestCase):

   def test_Static(self):
      self.assertEqual( starlink.Ast.tunec("HRDel"), "%-%^50+%s70+h%+" )
      starlink.Ast.tunec( "HRDel", "fred" )
      self.assertEqual( starlink.Ast.tunec("HRDel"), "fred" )
      self.assertEqual( starlink.Ast.escapes(1), 0 )
      self.assertEqual( starlink.Ast.escapes(0), 1 )
      self.assertEqual( starlink.Ast.escapes(-1), 0 )
      self.assertEqual( starlink.Ast.tune("ObjectCaching", 1 ), 0 )
      self.assertEqual( starlink.Ast.tune("ObjectCaching", starlink.Ast.TUNULL ), 1 )
      self.assertEqual( starlink.Ast.tune("ObjectCaching", 0 ), 1 )
      self.assertGreaterEqual( starlink.Ast.version(), 5007002 )
      self.assertTrue( os.path.isfile( os.path.join( starlink.Ast.get_include(),'star','pyast.h' ) ) )

   def test_Object(self):
      with self.assertRaises(TypeError):
         obj = starlink.Ast.Object()

   def test_ZoomMap(self):
      zoommap = starlink.Ast.ZoomMap( 1, 1.2 )
      self.assertEqual( zoommap.Class, "ZoomMap" )
      self.assertIsInstance( zoommap, starlink.Ast.ZoomMap )
      self.assertEqual( zoommap.Nobject, 1 )
      with self.assertRaises(AttributeError):
         zoommap.fred = 1.0
      with self.assertRaises(TypeError):
         zoommap.ID = 1.0
      zoommap.ID = "Hello there"
      self.assertEqual( zoommap.ID, "Hello there" )
      self.assertEqual( zoommap.get("ID"), "Hello there" )
      self.assertTrue( zoommap.test("ID") )
      zoommap.clear("ID")
      self.assertFalse( zoommap.test("ID") )
      self.assertEqual( zoommap.ID, "" )
      zoommap.set("ID=fred")
      self.assertEqual( zoommap.ID, "fred" )
      self.assertTrue( zoommap.UseDefs )
      self.assertEqual( zoommap.Nin, 1 )
      self.assertEqual( zoommap.Nout, 1 )
      self.assertEqual( zoommap.Zoom, 1.2 )
      zoommap.Zoom = -1.3
      self.assertEqual( zoommap.Zoom, -1.3 )
      self.assertEqual( zoommap.get("Zoom"), "-1.3" )
      zm = copy.deepcopy(zoommap)
      self.assertEqual( zoommap.Nobject, 2 )
      self.assertIsInstance( zm, starlink.Ast.ZoomMap )
      self.assertEqual( zm.Zoom, -1.3 )
      self.assertEqual( zoommap.Zoom, -1.3 )
      zm.Zoom = 3.0
      self.assertEqual( zm.Zoom, 3.0 )
      self.assertEqual( zoommap.Zoom, -1.3 )
      zm2 = zoommap.copy()
      self.assertEqual( zoommap.Nobject, 3 )
      self.assertIsInstance( zm2, starlink.Ast.ZoomMap )
      self.assertEqual( zm2.Zoom, -1.3 )
      self.assertEqual( zoommap.Zoom, -1.3 )
      zm2.Zoom = 3.0
      self.assertEqual( zm2.Zoom, 3.0 )
      self.assertEqual( zoommap.Zoom, -1.3 )
      zm2 = None
      self.assertEqual( zoommap.Nobject, 2 )
      self.assertTrue( zoommap.same(zoommap) )
      self.assertFalse( zoommap.same(zm) )
      del zm
      self.assertEqual( zoommap.Nobject, 1 )
      self.assertTrue( zoommap.hasattribute("ID") )
      self.assertFalse( zoommap.hasattribute("FID") )
      self.assertTrue( zoommap.isaobject() )
      self.assertTrue( zoommap.isamapping() )
      self.assertTrue( zoommap.isazoommap() )

      self.assertEqual( zoommap.RefCount, 1 )
      with self.assertRaises(AttributeError):
         zoommap.Nin = 3
      with self.assertRaises(starlink.Ast.AstError):
         zoommap = starlink.Ast.ZoomMap( 1, 0 )
      with self.assertRaises(starlink.Ast.ZOOMI):
         zoommap = starlink.Ast.ZoomMap( 1, 0 )
      zoommap = starlink.Ast.ZoomMap( 1, 12.25, "Zoom=2.1" )
      self.assertEqual( zoommap.Zoom, 2.1 )
      with self.assertRaises(starlink.Ast.NOWRT):
         zoommap = starlink.Ast.ZoomMap( 1, 12.25, "Nin=3" )

      self.assertTrue( starlink.Ast.ZoomMap( 1, 1.0 ).simplify().isaunitmap() )


      zoommap.lock(1)

   def test_FrameSimple(self):
      frame = starlink.Ast.Frame( 2, "label(1)=a b,label(2)=c d" )
      self.assertIsInstance( frame, starlink.Ast.Frame )
      self.assertEqual( frame.Nin, 2 )
      self.assertEqual( frame.Nout, 2 )
      testtitle = "Test Frame"
      frame.Title = testtitle
      self.assertEqual( frame.Title, testtitle)
      self.assertEqual( frame.get("Title"), testtitle)
      self.assertEqual( frame.Label_1, "a b" )
      self.assertEqual( frame.Label_2, "c d" )
      frame.Label_2 = "A new label"
      self.assertEqual( frame.Label_2, "A new label" )
      frame.Label_2 = None
      self.assertEqual( frame.Label_2, "Axis 2" )

      # Some methods
   def test_FrameAngle(self):
      frame = starlink.Ast.Frame(2)
      angle = frame.angle( [4,3], [0,0], [4,0] )
      self.assertEqual( angle, math.atan2(3,4) )

   def test_FrameAxis(self):
      frame = starlink.Ast.Frame(2)
      angle = frame.axangle( [0,0], [4,3], 1 )
      self.assertEqual( angle, -math.atan2(3,4) )
      distance = frame.axdistance( 1, 0, 4 )
      self.assertEqual( distance, 4 )
      axoffset = frame.axoffset( 1, 1, 4 )
      self.assertEqual( axoffset, 5 )

   def test_FrameDistance(self):
      frame = starlink.Ast.Frame(2)
      distance = frame.distance( [0,0],[4,3] )
      self.assertEqual( distance, 5 )

   def test_FrameFormat(self):
      frame = starlink.Ast.Frame(2)
      format = frame.format( 1, 55.270)
      self.assertEqual( format, "55.27" )

   def test_FrameIntersect(self):
      frame = starlink.Ast.Frame(2)
      cross = frame.intersect( [-1,1],[1,1],[0,0],[2,2] )
      self.assertEqual( cross[0], 1.0 )
      self.assertEqual( cross[1], 1.0 )

   def test_FrameMatchAxes(self):
      frame = starlink.Ast.Frame(2)
      frame2 = starlink.Ast.Frame( 3 )
      axes = frame.matchaxes( frame2 )
      self.assertEqual( axes[0], 1 )
      self.assertEqual( axes[1], 2 )
      self.assertEqual( axes[2], 0 )

   def test_FrameNorm(self):
      frame = starlink.Ast.Frame(2)
      coords = [3,2]
      ncoords = frame.norm( coords )
      self.assertEqual( ncoords[0], coords[0] )

   def test_FrameOffset(self):
      frame = starlink.Ast.Frame(2)
      point = frame.offset( [0,0], [4,3], 10 )
      self.assertEqual( point[0], 8 )
      self.assertEqual( point[1], 6 )
      direction, point = frame.offset2( [0,0], math.atan2(4,3), 10 )
      self.assertAlmostEqual( point[0], 8 )
      self.assertAlmostEqual( point[1], 6 )

   def test_FramePerm(self):
      frame = starlink.Ast.Frame(2)
      frame.permaxes( [2,1] )
      nframe,mapping = frame.pickaxes( [ 2 ] )
      self.assertEqual(nframe.Nin, 1 )
      self.assertIsInstance( frame, starlink.Ast.Frame )
      self.assertIsInstance( mapping, starlink.Ast.PermMap )
      self.assertEqual( mapping.Nin, 2 )
      self.assertEqual( mapping.Nout, 1 )

   def test_FrameConvert(self):
      frame = starlink.Ast.Frame(2)
      nframe = starlink.Ast.Frame( 2 )
      fset = frame.convert( nframe )
      self.assertIsInstance( fset, starlink.Ast.FrameSet )
      self.assertEqual( fset.Nframe, 2 )
      fset2 = fset.findframe( nframe )
      self.assertIsInstance( fset, starlink.Ast.FrameSet )

   def test_FrameResolve(self):
      frame = starlink.Ast.Frame(2)
      point4,d1,d2 = frame.resolve( [0,0], [3,3], [0,4] )
      self.assertAlmostEqual( d2, 0.0 )
      self.assertAlmostEqual( d1, math.sqrt(18) )

   def test_FrameUnformat(self):
      frame = starlink.Ast.Frame(2)
      nchars,value = frame.unformat( 1, "56.4 #" )
      self.assertEqual( nchars, 5 )
      self.assertEqual( value, 56.4 )

   def test_FrameActiveUnit(self):
      frame = starlink.Ast.Frame(2)
      self.assertFalse( frame.ActiveUnit )
      frame.ActiveUnit = True
      self.assertTrue( frame.ActiveUnit )

   def test_FrameSet(self):
      frame = starlink.Ast.Frame( 2 )
      frameset = starlink.Ast.FrameSet( frame )
      self.assertIsInstance( frameset, starlink.Ast.FrameSet )
      self.assertEqual( frameset.Nframe, 1 )
      frameset.addframe( 1, starlink.Ast.UnitMap( 2 ), starlink.Ast.Frame(2) )
      self.assertEqual( frameset.Nframe, 2 )
      frame2 = frameset.getframe( 1 )
      self.assertTrue( frame2.same(frame) )
      mapping = frameset.getmapping( 1, 2 )
      self.assertIsInstance( mapping, starlink.Ast.UnitMap )
      frameset.remapframe( 1, starlink.Ast.UnitMap(2) )
      frameset.removeframe( 1 )
      self.assertEqual( frameset.Nframe, 1 )

   def test_Mapping(self):
      with self.assertRaises(TypeError):
         mapping = starlink.Ast.Mapping()
      zoommap = starlink.Ast.ZoomMap( 1, 1.2 )
      map1,map2,series,invert1,invert2 = zoommap.decompose()
      self.assertIsInstance( map1, starlink.Ast.ZoomMap )
      self.assertEqual( map1.Zoom, 1.2 )
      self.assertIsNone( map2 )
      self.assertFalse( invert1 )
      self.assertFalse( invert2 )
      self.assertTrue( series )
      self.assertFalse( zoommap.Invert )
      zoommap.Invert = True
      self.assertTrue( zoommap.Invert )
      zoommap.invert()
      self.assertFalse( zoommap.Invert )
      self.assertTrue( zoommap.IsLinear )
      self.assertFalse( zoommap.IsSimple )
      self.assertFalse( zoommap.Report )
      self.assertTrue( zoommap.TranForward )
      self.assertTrue( zoommap.TranInverse )

      xin = numpy.linspace( -1, 1, 10 )
      xout = zoommap.tran( xin )
      d = (1.2*xin - xout)**2
      self.assertEqual( d.sum(), 0.0 )

      xa = [ 0., 1., 2., -1., -2., -3., 1., 2., 4., 5. ]
      zoommap.tran( xa, True, xout )
      d = (1.2*numpy.array( xa ) - xout)**2
      self.assertEqual( d.sum(), 0.0 )

      zoommap = starlink.Ast.ZoomMap( 3, 2.0 )
      pin = numpy.array( [[1.,2.,3], [0.,1.,2], [2.,3.,4]] )
      pout = zoommap.tran( pin, False )
      d = (0.5*pin - pout)**2
      self.assertEqual( d.sum(), 0.0 )

      zoommap = starlink.Ast.ZoomMap( 2, 2.0 )
      pout = zoommap.trangrid( [1,0], [3,2], 0.001, 100, True )
      answer = numpy.array( [[ 2., 4., 6., 2., 4., 6., 2., 4., 6.],
                             [ 0., 0., 0., 2., 2., 2., 4., 4., 4.]] )
      d = (answer - pout)**2
      self.assertEqual( d.sum(), 0.0 )

      islin,fit = zoommap.linearapprox(  [1,0], [3,2], 0.001 )
      answer = numpy.array( [ 0., 0., 2., 0., 0., 2.] )
      d = (answer - fit)**2
      self.assertEqual( d.sum(), 0.0 )
      self.assertTrue( islin )

      lb,ub,xl,xu = zoommap.mapbox(  [1,0], [3,2], 2 )
      self.assertEqual( lb, 0 )
      self.assertEqual( ub, 4 )
      self.assertEqual( xl[1], 0 )
      self.assertEqual( xu[1], 2 )

      isquad,fit,rms = zoommap.quadapprox(  [1,0], [3,2], 3, 3 )

      self.assertTrue( isquad )
      self.assertEqual( rms, 0.0 )
      answer = numpy.array( [ 0.,2.,0.,0.,0.,0.,0.,0.,2.,0.,0.,0.] )
      d = (answer - fit)**2
      self.assertEqual( d.sum(), 0.0 )

      self.assertEqual( zoommap.rate( [1,1], 2, 2 ), 2.0 )
      self.assertEqual( zoommap.rate( [1,1], 1, 2 ), 0.0 )

      data_in = numpy.array( [[ 1., 2., 3.,],
                              [ 4., 5., 6.],
                              [ 7., 8., 9.] ] )
      zoommap = starlink.Ast.ZoomMap( 2, 1.0 )
      out,outv = zoommap.rebin( 0.5, [1,0], [3,2], data_in, None,
                                starlink.Ast.NEAREST, None, starlink.Ast.USEBAD,
                                0.0, 100, starlink.Ast.BAD, [2,0], [4,2], [1,0],
                                [3,2] )

      answer = numpy.array( [ [2., 3., starlink.Ast.BAD],
                              [5., 6., starlink.Ast.BAD],
                              [8., 9., starlink.Ast.BAD]] )
      d = (answer - out)**2
      self.assertEqual( d.sum(), 0.0 )
      self.assertIsNone( outv )

      data_in = numpy.array( [[1,2,3],[4,5,6],[7,8,9]], dtype=numpy.intc )
      data_out = numpy.empty( (3,3), dtype=numpy.intc )
      weights = numpy.zeros( (3,3), dtype=numpy.double )

      flags = starlink.Ast.USEBAD | starlink.Ast.REBININIT
      nused = 0
      nused = zoommap.rebinseq( 0.5, [1,0], [3,2], data_in, None,
                                starlink.Ast.LINEAR, None, flags,
                                0.0, 100, -999, [2,0], [4,2],
                                [1,0], [3,2], data_out, None, weights,
                                nused )
      flags = starlink.Ast.USEBAD | starlink.Ast.REBINEND
      nused = zoommap.rebinseq( 0.5, [1,0], [3,2], data_in, None,
                                starlink.Ast.LINEAR, None, flags,
                                0.0, 100, -999, [2,0], [4,2],
                                [1,0], [3,2], data_out, None, weights,
                                nused )


      answer = numpy.array( [[ 2., 3., -999],
                             [ 5., 6., -999],
                             [ 8., 9., -999]] )
      d = (answer - data_out)**2
      self.assertEqual( d.sum(), 0.0 )
      self.assertEqual( nused, 12 )

      data_in = numpy.array( [[ 1., 2., 3.,],
                              [ 4., 5., 6.],
                              [ 7., 8., 9.] ] )

      zoommap = starlink.Ast.ZoomMap( 2, 1.0 )
      npix,out,outv = zoommap.resample( [1,0], [3,2], data_in, None,
                                   starlink.Ast.NEAREST, None, starlink.Ast.USEBAD,
                                   0.0, 100, starlink.Ast.BAD, [2,0], [4,2], [2,0],
                                   [4,2] )

      answer = numpy.array( [[ 2., 3., starlink.Ast.BAD],
                             [ 5., 6., starlink.Ast.BAD],
                             [ 8., 9., starlink.Ast.BAD] ] )
      d = (answer - out)**2
      self.assertEqual( d.sum(), 0.0 )
      self.assertIsNone( outv )
      self.assertEqual( npix, 3 )



   def test_PermMap(self):
      permmap = starlink.Ast.PermMap( [2,1,3],[1,2] )
      self.assertIsInstance( permmap, starlink.Ast.PermMap )
      self.assertEqual( permmap.Nin, 3 )
      self.assertEqual( permmap.Nout, 2 )

   def test_ShiftMap(self):
      shiftmap = starlink.Ast.ShiftMap( [1,2] )
      self.assertIsInstance( shiftmap, starlink.Ast.ShiftMap )
      self.assertIsInstance( shiftmap, starlink.Ast.Mapping )
      self.assertEqual( shiftmap.Nout, 2 )

   def test_SphMap(self):
      sphmap = starlink.Ast.SphMap( "UnitRadius=1" )
      self.assertIsInstance( sphmap, starlink.Ast.SphMap )
      self.assertIsInstance( sphmap, starlink.Ast.Mapping )
      self.assertEqual( sphmap.Nout, 2 )
      self.assertTrue( sphmap.UnitRadius )

   def test_RateMap(self):
      ratemap = starlink.Ast.RateMap( starlink.Ast.UnitMap(2) )
      self.assertIsInstance( ratemap, starlink.Ast.RateMap )
      self.assertIsInstance( ratemap, starlink.Ast.Mapping )
      self.assertEqual( ratemap.Nout, 1 )

   def test_WcsMap(self):
      wcsmap = starlink.Ast.WcsMap()
      self.assertIsInstance( wcsmap, starlink.Ast.WcsMap )
      self.assertIsInstance( wcsmap, starlink.Ast.Mapping )
      self.assertEqual( wcsmap.Nout,2 )
      with self.assertRaises(starlink.Ast.WCSTY):
         wcsmap = starlink.Ast.WcsMap( 2, 4000, 1, 2 )
      wcsmap.set( "PV2_0=1.2" )
      self.assertEqual( wcsmap.ProjP_0, 1.2 )
      self.assertEqual( wcsmap.WcsType, starlink.Ast.TAN )

   def test_PcdMap(self):
      pcdmap = starlink.Ast.PcdMap( 0.2, [1,2] )
      self.assertIsInstance( pcdmap, starlink.Ast.PcdMap )
      self.assertIsInstance( pcdmap, starlink.Ast.Mapping )
      self.assertEqual( pcdmap.Nout, 2 )
      self.assertEqual( pcdmap.Disco, 0.2 )
      with self.assertRaises(ValueError):
         pcdmap = starlink.Ast.PcdMap( 0.5, [1,2,3])

   def test_LutMap(self):
      lutmap = starlink.Ast.LutMap( [1,2,3,4,5], 1, 0.5 )
      self.assertIsInstance( lutmap, starlink.Ast.LutMap )
      self.assertIsInstance( lutmap, starlink.Ast.Mapping )
      self.assertEqual( lutmap.Nout, 1 )

   def test_UnitMap(self):
      unitmap = starlink.Ast.UnitMap( 3 )
      self.assertIsInstance( unitmap, starlink.Ast.UnitMap )
      self.assertEqual( unitmap.Nin, 3 )

   def test_GrismMap(self):
      grismmap = starlink.Ast.GrismMap("GrismM=1")
      self.assertIsInstance( grismmap, starlink.Ast.GrismMap )
      self.assertIsInstance( grismmap, starlink.Ast.Mapping )
      self.assertEqual( grismmap.GrismM, 1 )

   def test_WinMap(self):
      with self.assertRaises(ValueError):
         winmap = starlink.Ast.WinMap( [1], [1,2], [3,4],[5,6])
      winmap = starlink.Ast.WinMap( [1,1],[2,2], [1,1],[3,3] )
      self.assertIsInstance( winmap, starlink.Ast.WinMap )
      self.assertEqual( winmap.Nin, 2 )

   def test_NormMap(self):
      with self.assertRaises(TypeError):
         normmap = starlink.Ast.NormMap( starlink.Ast.UnitMap( 1 ) )
      normmap = starlink.Ast.NormMap( starlink.Ast.Frame( 2 ) )
      self.assertIsInstance( normmap, starlink.Ast.NormMap )
      self.assertEqual( normmap.Nin, 2 )

   def test_TimeMap(self):
      with self.assertRaises(ValueError):
         timemap = starlink.Ast.TimeMap( 1 )
      timemap = starlink.Ast.TimeMap( )
      self.assertIsInstance( timemap, starlink.Ast.TimeMap )
      self.assertIsInstance( timemap, starlink.Ast.Mapping )
      self.assertEqual( timemap.Nin, 1 )
      timemap.timeadd( "BEPTOMJD", [560,720] )
      with self.assertRaises(starlink.Ast.TIMIN):
         timemap.timeadd( "UNRECOGNIZED", [1] )

   def test_CmpFrame(self):
      cmpframe = starlink.Ast.CmpFrame( starlink.Ast.Frame(2), starlink.Ast.Frame(2) )
      self.assertIsInstance( cmpframe, starlink.Ast.CmpFrame )
      self.assertIsInstance( cmpframe, starlink.Ast.Frame )

   def test_CmpMap(self):
      cmpmap = starlink.Ast.CmpMap( starlink.Ast.UnitMap(2), starlink.Ast.UnitMap(2), True )
      self.assertIsInstance( cmpmap, starlink.Ast.CmpMap )
      self.assertIsInstance( cmpmap, starlink.Ast.Mapping )

      zoommap = starlink.Ast.ZoomMap( 2, 1.0 )
      unitmap = starlink.Ast.UnitMap( 1 )
      cmpmap = starlink.Ast.CmpMap( zoommap, unitmap, False )
      out,map = cmpmap.mapsplit( 3 )

      self.assertIsInstance( map, starlink.Ast.UnitMap )
      self.assertEqual( len(out), 3 )
      self.assertEqual( out[0], 3 )
      self.assertEqual( out[1], 0 )
      self.assertEqual( out[2], 0 )

   def test_TranMap(self):
      tranmap = starlink.Ast.TranMap( starlink.Ast.UnitMap(2), starlink.Ast.UnitMap(2) )
      self.assertIsInstance( tranmap, starlink.Ast.TranMap )
      self.assertIsInstance( tranmap, starlink.Ast.Mapping )
      with self.assertRaises(TypeError):
         tranmap = starlink.Ast.TranMap( starlink.Ast.UnitMap( 1 ), [1] )

   def test_SpecFrame(self):
      specframe = starlink.Ast.SpecFrame()
      self.assertIsInstance( specframe, starlink.Ast.Frame )
      self.assertIsInstance( specframe, starlink.Ast.SpecFrame )
      sky = starlink.Ast.SkyFrame()
      specframe.setrefpos( sky, 0, 1 )
      refpos = specframe.getrefpos(sky)
      self.assertAlmostEqual( refpos[0], 0, 5 )
      self.assertAlmostEqual( refpos[1], 1 )

   def test_DSBSpecFrame(self):
      dsbspecframe = starlink.Ast.DSBSpecFrame( "IF=4.0,AlignSideBand=1")
      self.assertIsInstance( dsbspecframe, starlink.Ast.Frame )
      self.assertIsInstance( dsbspecframe, starlink.Ast.SpecFrame )
      self.assertIsInstance( dsbspecframe, starlink.Ast.DSBSpecFrame )
      self.assertTrue( dsbspecframe.AlignSideBand )
      self.assertEqual( dsbspecframe.IF, 4.0 )

   def test_SkyFrame(self):
      skyframe = starlink.Ast.SkyFrame()
      self.assertIsInstance( skyframe, starlink.Ast.Frame )
      self.assertIsInstance( skyframe, starlink.Ast.SkyFrame )
      mapping = skyframe.skyoffsetmap()
      self.assertIsInstance( mapping, starlink.Ast.Mapping )

   def test_TimeFrame(self):
      tframe = starlink.Ast.TimeFrame( "TimeScale=TAI" )
      self.assertIsInstance( tframe, starlink.Ast.Frame )
      self.assertEqual( tframe.TimeScale, "TAI" )
      self.assertGreater( tframe.currenttime(), 55700.0)

   def test_FluxFrame(self):
      fframe = starlink.Ast.FluxFrame()
      self.assertIsInstance( fframe, starlink.Ast.Frame )
      self.assertIsInstance( fframe, starlink.Ast.FluxFrame )

   def test_SpecFluxFrame(self):
      sfframe = starlink.Ast.SpecFluxFrame( starlink.Ast.SpecFrame(),
                                            starlink.Ast.FluxFrame(57, starlink.Ast.SpecFrame()))
      self.assertIsInstance( sfframe, starlink.Ast.Frame )
      self.assertIsInstance( sfframe, starlink.Ast.CmpFrame )
      self.assertIsInstance( sfframe, starlink.Ast.SpecFluxFrame )

   def test_Box(self):
      box = starlink.Ast.Box( starlink.Ast.Frame(2), 1,
                             [0,0],[3,4] )
      self.assertIsInstance( box, starlink.Ast.Box )
      self.assertIsInstance( box, starlink.Ast.Region )
      self.assertIsInstance( box, starlink.Ast.Frame )
      self.assertIsInstance( box.getregionframe(), starlink.Ast.Frame )
      lbnd,ubnd = box.getregionbounds()
      self.assertEqual( lbnd[0], 0 )
      self.assertEqual( ubnd[0], 3 )
      testbox = starlink.Ast.Box( starlink.Ast.Frame(2), 1,
                                 [1,1],[4,5])
      overlap = box.overlap( testbox )
      self.assertEqual( overlap, 4 )
      self.assertTrue(  box.removeregions().isaframe() )

   def test_MapRegion(self):
      box = starlink.Ast.Box( starlink.Ast.Frame(2), 1,
                             [0,0],[3,4] )
      frame = starlink.Ast.Frame(2)
      mapping = starlink.Ast.ZoomMap( 2, 2.0 )
      region = box.mapregion( mapping, frame )
      lbnd,ubnd = region.getregionbounds()
      self.assertEqual( lbnd[0], 0 )
      self.assertEqual( ubnd[0], 6 )
      self.assertEqual( lbnd[1], 0 )
      self.assertEqual( ubnd[1], 8 )

   def test_Circle(self):
      circle = starlink.Ast.Circle( starlink.Ast.Frame(2), 0,
                             [0,0],[3,4] )
      self.assertIsInstance( circle, starlink.Ast.Circle )
      self.assertIsInstance( circle, starlink.Ast.Region )
      self.assertIsInstance( circle, starlink.Ast.Frame )
      self.assertIsInstance( circle.getregionframe(), starlink.Ast.Frame )
      testcircle = starlink.Ast.Circle( starlink.Ast.Frame(2), 1,
                                 [0,0],5)
      overlap = circle.overlap( testcircle )
      self.assertEqual( overlap, 5 )

   def test_Ellipse(self):
      ell = starlink.Ast.Ellipse( starlink.Ast.Frame(2), 0,
                             [0,0],[3,4], [2,3] )
      self.assertIsInstance( ell, starlink.Ast.Ellipse )
      self.assertIsInstance( ell, starlink.Ast.Region )
      self.assertIsInstance( ell, starlink.Ast.Frame )

   def test_NullRegion(self):
      null = starlink.Ast.NullRegion( starlink.Ast.Frame(2) )
      self.assertIsInstance( null, starlink.Ast.NullRegion )
      self.assertIsInstance( null, starlink.Ast.Region )
      self.assertFalse( null.Negated )

   def test_Interval(self):
      interval = starlink.Ast.Interval( starlink.Ast.Frame(2), [1,2],[5,10] )
      self.assertIsInstance( interval, starlink.Ast.Interval )
      self.assertIsInstance( interval, starlink.Ast.Region )

   def test_CmpRegion(self):
      circle = starlink.Ast.Circle( starlink.Ast.Frame(2), 0,
                             [0,0],[3,4] )
      box = starlink.Ast.Box( starlink.Ast.Frame(2), 1,
                              [0,0],[3,4] )
      cmp = starlink.Ast.CmpRegion( circle, box, starlink.Ast.AND )
      self.assertIsInstance( cmp, starlink.Ast.CmpRegion )
      self.assertIsInstance( cmp, starlink.Ast.Region )

   def test_Prism(self):
      circle = starlink.Ast.Circle( starlink.Ast.Frame(2), 0,
                             [0,0],[3,4] )
      interval = starlink.Ast.Interval( starlink.Ast.Frame(1), [6],[100] )
      prism = starlink.Ast.Prism( circle, interval )
      self.assertIsInstance( prism, starlink.Ast.Prism )
      self.assertIsInstance( prism, starlink.Ast.Region )

   def test_Channel(self):
      channel = starlink.Ast.Channel()
      self.assertIsInstance( channel, starlink.Ast.Object)
      self.assertIsInstance( channel, starlink.Ast.Channel )
      self.assertTrue(  channel.isachannel() )
      self.assertTrue(  channel.isaobject() )
      zoommap = starlink.Ast.ZoomMap( 2, 0.1, "ID=Hello there" )
      channel.SinkFile= "fred.txt"
      n = channel.write( zoommap )
      self.assertEqual( n, 1 )
      channel.SourceFile= "fred.txt"
      with self.assertRaises(starlink.Ast.RDERR):
         obj = channel.read( )
      channel.SinkFile = None
      channel.SourceFile= "fred.txt"
      obj = channel.read( )
      channel.SinkFile = "fred2.txt"
      channel.write( obj )
      channel.SinkFile = None
      channel.SourceFile = None
      self.assertTrue( filecmp.cmp("fred.txt", "fred2.txt", shallow=False)  )
      os.remove("fred.txt")
      os.remove("fred2.txt")

   def test_MyChannel(self):
      ss = DummyStream()
      with self.assertRaises(TypeError):
         channel = starlink.Ast.Channel( ss, ss )
      ss = TextStream()
      channel = starlink.Ast.Channel( ss, ss )
      zoommap = starlink.Ast.ZoomMap( 2, 0.1, "ID=Hello there" )
      n = channel.write( zoommap )
      self.assertEqual( n, 1 )
      a = ss.get()
      obj = channel.read( )
      ss.reset_sink()
      ss.reset_source()
      n = channel.write( obj )
      self.assertEqual( n, 1 )
      b = ss.get()
      self.assertEqual( a, b )

   def test_FitsChan(self):
      mycards = ("NAXIS1  =                  200                                                  ",
                 "NAXIS2  =                  200                                                  ",
                 "CTYPE1  = 'RA--TAN '                                                            ",
                 "CTYPE2  = 'DEC-TAN '                                                            ",
                 "CRPIX1  =                  100                                                  ",
                 "CRPIX2  =                  100                                                  ",
                 "CDELT1  =                0.001                                                  ",
                 "CDELT2  =                0.001                                                  ",
                 "CRVAL1  =                    0                                                  ",
                 "CRVAL2  =                    0                                                  ")

      fc = starlink.Ast.FitsChan(mycards)
      self.assertIsInstance( fc, starlink.Ast.Object)
      self.assertEqual( fc[ "CRVAL1" ], 0.0 )

      fc = starlink.Ast.FitsChan()
      self.assertIsInstance( fc, starlink.Ast.Object)
      self.assertIsInstance( fc, starlink.Ast.Channel )
      self.assertIsInstance( fc, starlink.Ast.FitsChan )
      self.assertTrue( fc.isafitschan() )
      self.assertTrue( fc.isachannel() )
      self.assertTrue( fc.isaobject() )

      fc.setfitsI( "FRED", 99, "Hello there", True)
      there,value = fc.getfitsI( "FRED" )
      self.assertTrue( there )
      self.assertEqual( value, 99 )
      there,value = fc.getfitsS( "FRED" )
      self.assertTrue( there )
      self.assertEqual( value, "99" )
      fc.setfitsF( "FRED1", 99.9, "Hello there", True)
      there,value = fc.getfitsS( "FRED1" )
      self.assertTrue( there )
      self.assertEqual( value, "99.9" )
      fc.setfitsCF( "FRED2", (99.9,99.8), "Hello there", True)
      there,(real,imag) = fc.getfitsCF( "FRED2" )
      self.assertTrue( there )
      self.assertEqual( real, 99.9 )
      self.assertEqual( imag, 99.8 )
      fc.setfitsCI( "FRED3", (99,98), "Hello there", True)
      there,value = fc.getfitsS( "FRED3" )
      self.assertTrue( there )
      self.assertEqual( value, "99 98" )
      fc.setfitsS( "FRED4", "-12", "Hello there", True)
      there,value = fc.getfitsI( "FRED4" )
      self.assertTrue( there )
      self.assertEqual( value, -12 )
      fc.emptyfits()
      self.assertEqual( fc.Ncard, 0 )

      cards = "CRVAL1  = 0                                                                     CRVAL2  = 0                                                                     "
      fc.putcards( cards )
      self.assertEqual( fc.Card, 1 )
      fc.set( "Card=10" )
      self.assertEqual( fc.Card, 3 )
      fc.Card = None
      self.assertEqual( fc.Card, 1 )

      for card in mycards[0:8]:
         fc.putfits( card, False )

      self.assertEqual( fc.Ncard, 10 )
      self.assertEqual( fc.Card, 9 )
      self.assertEqual( fc.Encoding, "FITS-WCS" )
      there, card = fc.findfits( "%f", False )
      self.assertTrue( there )
      self.assertEqual( card, "CRVAL1  =                    0                                                  " )
      fc.delfits()
      self.assertEqual( fc.Ncard, 9 )
      self.assertEqual( fc.Card, 9 )
      there, card = fc.findfits( "%f", False )
      self.assertTrue( there )
      self.assertEqual( card, "CRVAL2  =                    0                                                  " )
      fc.putfits( "CRVAL1  = 0", False )
      self.assertEqual( fc.Ncard, 10 )
      self.assertEqual( fc.Card, 10 )
      there, card = fc.findfits( "%f", False )
      self.assertTrue( there )
      self.assertEqual( card, "CRVAL2  =                    0                                                  " )

      for cards in zip(fc,mycards):
         self.assertEqual( cards[0], cards[1] )

      obj = fc.read()
      self.assertIsInstance( obj, starlink.Ast.FrameSet)
      self.assertTrue( obj.isaframeset() )

      with self.assertRaises(TypeError):
         fc = starlink.Ast.FitsChan("Encoding=FITS-WCS")

      fc = starlink.Ast.FitsChan( None, None, "Encoding=FITS-WCS")

      n = fc.write( obj )
      self.assertEqual( n, 1 )
      fc.Card = None

      there, card = fc.findfits( "%f", True )
      card = card.rstrip()
      self.assertTrue( there )
      self.assertEqual(card, "WCSAXES =                    2 / Number of WCS axes" )

      there, card = fc.findfits( "%f", True )
      card = card.rstrip()
      self.assertTrue( there )
      self.assertEqual(card, "CRPIX1  =                100.0 / Reference pixel on axis 1" )

      there, card = fc.findfits( "%f", True )
      card = card.rstrip()
      self.assertTrue( there )
      self.assertEqual(card, "CRPIX2  =                100.0 / Reference pixel on axis 2" )

      there, card = fc.findfits( "%f", True )
      card = card.rstrip()
      self.assertTrue( there )
      self.assertEqual(card, "CRVAL1  =                  0.0 / Value at ref. pixel on axis 1" )

      there, card = fc.findfits( "%f", True )
      card = card.rstrip()
      self.assertTrue( there )
      self.assertEqual(card, "CRVAL2  =                  0.0 / Value at ref. pixel on axis 2" )

      there, card = fc.findfits( "%f", True )
      card = card.rstrip()
      self.assertTrue( there )
      self.assertEqual(card, "CTYPE1  = 'RA---TAN'           / Type of co-ordinate on axis 1" )

      there, card = fc.findfits( "%f", True )
      card = card.rstrip()
      self.assertTrue( there )
      self.assertEqual(card, "CTYPE2  = 'DEC--TAN'           / Type of co-ordinate on axis 2" )

      there, card = fc.findfits( "%f", True )
      card = card.rstrip()
      self.assertTrue( there )
      self.assertEqual(card, "CDELT1  =                0.001 / Pixel size on axis 1" )

      there, card = fc.findfits( "%f", True )
      card = card.rstrip()
      self.assertTrue( there )
      self.assertEqual(card, "CDELT2  =                0.001 / Pixel size on axis 2" )

      there, card = fc.findfits( "%f", True )
      card = card.rstrip()
      self.assertTrue( there )
      self.assertEqual(card, "RADESYS = 'ICRS    '           / Reference frame for RA/DEC values" )

      ss = TextStream()
      fc = starlink.Ast.FitsChan( ss, ss, "Encoding=FITS-WCS")
      n = fc.write( obj )
      self.assertEqual( n, 1 )
      fc.writefits()
      a = ss.get()
      fc.readfits()
      obj = fc.read( )
      ss.reset_sink()
      ss.reset_source()
      n = fc.write( obj )
      self.assertEqual( n, 1 )
      fc = None
      b = ss.get()
      self.assertEqual( a, b )

   def test_FitsChan_AsMapping(self):
      fc = starlink.Ast.FitsChan()
      fc["NAXIS1"] = 200
      fc["NAXIS2"] = 200
      fc["CTYPE1"] = "RA--TAN "
      fc["CTYPE2"] = "DEC-TAN "
      fc["CRPIX1"] = 100
      fc["CRPIX2"] = 100
      fc["CDELT1"] = 0.001
      fc["CDELT2"] = 0.001
      fc["CRVAL1"] = 0
      fc["CRVAL2"] = 0

      self.assertTrue( "NAXIS1" in fc )

      self.assertEqual( len(fc), 10 )
      self.assertEqual( fc[2], "CTYPE1  = 'RA--TAN '                                                            ")
      obj = fc.read( )
      self.assertEqual( len(fc), 2 )

      with self.assertRaises(KeyError):
         a = fc["CRVAL1"]

      self.assertEqual( fc["NAXIS1"], 200 )
      fc["NAXIS2"] = None
      self.assertEqual( len(fc), 1 )
      fc["NAXIS2"] = "This is a string"
      self.assertEqual( len(fc), 2 )
      self.assertEqual( fc["NAXIS2"], "This is a string")
      fc.setfitsI( "NAXIS2", -12, "Hello there", False )
      self.assertEqual( fc.Ncard, 3 )
      self.assertEqual( fc.Nkey, 2 )
      self.assertEqual( len(fc), 2 )
      self.assertEqual( fc["NAXIS2"][0], -12 )
      self.assertEqual( fc["NAXIS2"][1], "This is a string" )
      fc["NAXIS2"] = 1000
      self.assertEqual( fc["NAXIS2"], 1000 )
      self.assertEqual( fc.Ncard, 2 )
      self.assertEqual( fc.Nkey, 2 )
      self.assertEqual( len(fc), 2 )

      fc[100] = "CTYPE1  = 'RA--TAN '                                                            "
      self.assertEqual( fc[2], "CTYPE1  = 'RA--TAN '                                                            ")
      self.assertEqual( fc.Ncard, 3 )
      self.assertEqual( fc.Nkey, 3 )
      self.assertEqual( len(fc), 3 )

      fc[0] = "NEWKEY  = 123.456"
      self.assertEqual( fc[0], "NEWKEY  =              123.456                                                  ")
      self.assertEqual( fc.Ncard, 3 )
      self.assertEqual( fc.Nkey, 3 )
      self.assertEqual( len(fc), 3 )

   def test_StcsChan(self):
      ss = TextStream()
      ch = starlink.Ast.StcsChan( ss, ss, "ReportLevel=3" )
      self.assertIsInstance( ch, starlink.Ast.Object)
      self.assertIsInstance( ch, starlink.Ast.Channel )
      self.assertIsInstance( ch, starlink.Ast.StcsChan )
      self.assertTrue( ch.isastcschan() )
      self.assertTrue( ch.isachannel() )
      self.assertTrue( ch.isaobject() )
      ss.astsink("StartTime 1900-01-01 Circle ICRS 148.9 69.1 2.0")
      ss.astsink("SpeCtralInterval 4000 7000 unit Angstrom")
      obj = ch.read()
      self.assertIsInstance( obj, starlink.Ast.Prism )
      self.assertEqual( obj.Class, "Prism" )
      self.assertEqual( obj.Naxes, 4 )
      lbnd,ubnd = obj.getregionbounds()
      self.assertEqual( ubnd[0], sys.float_info.max )
      self.assertAlmostEqual( lbnd[1], 2.50080939227851 )
      self.assertAlmostEqual( ubnd[1], 2.6967811201606 )
      self.assertAlmostEqual( lbnd[2], 1.171115928088195 )
      self.assertAlmostEqual( ubnd[2], 1.24091013301998 )
      self.assertEqual( lbnd[3], 4000.0 )
      self.assertEqual( ubnd[3], 7000.0 )

   def test_KeyMap(self):

      with self.assertRaises(starlink.Ast.BADAT):
         km = starlink.Ast.KeyMap( "ReportLevel=3" )

      km = starlink.Ast.KeyMap( "KeyCase=0" )
      self.assertIsInstance( km, starlink.Ast.Object)
      self.assertIsInstance( km, starlink.Ast.KeyMap )
      self.assertTrue( km.isakeymap() )
      self.assertTrue( km.isaobject() )
      self.assertEqual( len(km), 0 )
      km['Hello'] = 'fred'
      self.assertEqual( len(km), 1 )
      self.assertEqual( km['hELlo'], 'fred' )
      km['Hello'] = None
      self.assertEqual( len(km), 0 )
      km['TTT'] = (1.2, 3, 4.5)
      self.assertEqual( len(km), 1 )
      self.assertEqual( km['ttt'], (1.2, 3, 4.5) )
      km['SS'] = 'hello'
      self.assertEqual( len(km), 2 )

      self.assertEqual( km[0], ('TTT', (1.2, 3.0, 4.5)) )
      self.assertEqual( km[1], ('SS', 'hello'))
      self.assertTrue( "SS" in km )

      i = 0
      for entry in km:
         self.assertEqual( entry, km[i] )
         i += 1

      km[0] = None
      self.assertEqual( len(km), 1 )
      self.assertEqual( km[0], ('SS', 'hello'))
      km[0] = 'Goofbye'
      self.assertEqual( km[0], ('SS', 'Goofbye'))
      with self.assertRaises(starlink.Ast.MPIND):
         km[1] = 'Nooooooo'

   def test_Plot(self):
      with self.assertRaises(TypeError):
         plot = starlink.Ast.Plot( None, [0.0, 0.0, 1.0, 1.0], [0.0, 0.0, 1.0, 1.0],
                                   TextStream() )
      mygrf = DummyGrf()
      plot = starlink.Ast.Plot( None, [0.0, 0.0, 1.0, 1.0], [0.0, 0.0, 1.0, 1.0],
                                mygrf )
      self.assertIsInstance( plot, starlink.Ast.Object)
      self.assertIsInstance( plot, starlink.Ast.Mapping )
      self.assertIsInstance( plot, starlink.Ast.Frame )
      self.assertIsInstance( plot, starlink.Ast.FrameSet )
      self.assertIsInstance( plot, starlink.Ast.Plot )
      self.assertTrue( plot.isaplot() )
      self.assertTrue( plot.isaframeset() )
      self.assertTrue( plot.isaframe() )
      self.assertTrue( plot.isamapping() )
      self.assertTrue( plot.isaobject() )

      self.assertEqual( plot.Tol, 0.01 )
      plot.Tol = 0.5
      self.assertEqual( plot.Tol, 0.5 )

      self.assertEqual( plot.Colour_Border, 1 )
      plot.Colour_Border = 2
      self.assertEqual( plot.Colour_Border, 2 )

      plot.border()

      self.assertAlmostEqual( mygrf.linex, [0.0, 0.071428574621677399, 0.1428571492433548, 0.2142857164144516, 0.28571429848670959, 0.3571428656578064, 0.4285714328289032, 0.5, 0.57142859697341919, 0.6428571343421936, 0.71428573131561279, 0.78571426868438721, 0.8571428656578064, 0.92857140302658081, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.92857140302658081, 0.8571428656578064, 0.78571426868438721, 0.71428573131561279, 0.6428571343421936, 0.57142859697341919, 0.5, 0.4285714328289032, 0.3571428656578064, 0.28571429848670959, 0.2142857164144516, 0.1428571492433548, 0.071428574621677399, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0] )
      self.assertAlmostEqual( mygrf.liney, [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.071428574621677399, 0.1428571492433548, 0.2142857164144516, 0.28571429848670959, 0.3571428656578064, 0.4285714328289032, 0.5, 0.57142859697341919, 0.6428571343421936, 0.71428573131561279, 0.78571426868438721, 0.8571428656578064, 0.92857140302658081, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.92857140302658081, 0.8571428656578064, 0.78571426868438721, 0.71428573131561279, 0.6428571343421936, 0.57142859697341919, 0.5, 0.4285714328289032, 0.3571428656578064, 0.28571429848670959, 0.2142857164144516, 0.1428571492433548, 0.071428574621677399, 0.0] )
      self.assertEqual( mygrf.nline, 57 )
      self.assertEqual( mygrf.attr, [4, 4] )
      self.assertAlmostEqual( mygrf.value, [2.0, 1.0] )
      self.assertEqual( mygrf.prim, [1, 1] )

      mygrf.Reset()
      plot.grid()
      self.assertEqual( mygrf.textt, ['0', '0.3', '0.6', '0.9', '0.4', '0.7', '1', 'Axis 1', 'Axis 2', '2-d coordinate system'])
      self.assertEqual( mygrf.textj,  ['TC', 'TC', 'TC', 'TC', 'CR', 'CR', 'CR', 'TC', 'BC', 'BC'])
      self.assertEqual( mygrf.ntext, 10)

      mygrf.Reset()
      pin = numpy.array( [[0.5,1.,0.], [0.5,1.,0.5]] )
      plot.polycurve( pin )
      self.assertEqual( mygrf.nline, 29 )

      lbnd,ubnd = plot.boundingbox()
      self.assertAlmostEqual( lbnd[0], 0 )
      self.assertAlmostEqual( lbnd[1], 0.5 )
      self.assertAlmostEqual( ubnd[0], 1. )
      self.assertAlmostEqual( ubnd[1], 1. )

      plot.Tol = 0.0000001
      plot.clip( starlink.Ast.CURRENT, [0.1,0.1], [0.9,0.9] )
      plot.polycurve( pin )
      lbnd,ubnd = plot.boundingbox()
      self.assertAlmostEqual( lbnd[0], 0.1 )
      self.assertAlmostEqual( lbnd[1], 0.5 )
      self.assertAlmostEqual( ubnd[0], 0.9 )
      self.assertAlmostEqual( ubnd[1], 0.9 )

      plot.curve( [0.0,0.0], [1.0,1.0] )
      lbnd,ubnd = plot.boundingbox()
      self.assertAlmostEqual( lbnd[0], 0.1 )
      self.assertAlmostEqual( lbnd[1], 0.1 )
      self.assertAlmostEqual( ubnd[0], 0.9 )
      self.assertAlmostEqual( ubnd[1], 0.9 )

      plot.clip( starlink.Ast.NOFRAME, [], [] )
      plot.curve( [0.0,0.0], [1.0,1.0] )
      lbnd,ubnd = plot.boundingbox()
      self.assertAlmostEqual( lbnd[0], 0.0 )
      self.assertAlmostEqual( lbnd[1], 0.0 )
      self.assertAlmostEqual( ubnd[0], 1.0 )
      self.assertAlmostEqual( ubnd[1], 1.0 )

      xlut = starlink.Ast.LutMap( [0,1], 0, 1.0 )
      ylut = starlink.Ast.LutMap( [0,0.5], 0, 1.0 )
      xylut = starlink.Ast.CmpMap( xlut, ylut, False )
      pm = starlink.Ast.PermMap( [1], [1,1] )
      map = starlink.Ast.CmpMap( pm, xylut, True )
      plot.gencurve( map )
      lbnd,ubnd = plot.boundingbox()
      self.assertAlmostEqual( lbnd[0], 0.0 )
      self.assertAlmostEqual( lbnd[1], 0.0 )
      self.assertAlmostEqual( ubnd[0], 1.0 )
      self.assertAlmostEqual( ubnd[1], 0.5 )

      plot.gridline( 1, [0.5,0.5], 0.5 )
      lbnd,ubnd = plot.boundingbox()
      self.assertAlmostEqual( lbnd[0], 0.5 )
      self.assertAlmostEqual( lbnd[1], 0.5 )
      self.assertAlmostEqual( ubnd[0], 1.0 )
      self.assertAlmostEqual( ubnd[1], 0.5 )

      mygrf.Reset()
      plot.mark( pin, 1 )
      lbnd,ubnd = plot.boundingbox()
      self.assertAlmostEqual( lbnd[0], 0.0 )
      self.assertAlmostEqual( lbnd[1], 0.5 )
      self.assertAlmostEqual( ubnd[0], 1.0 )
      self.assertAlmostEqual( ubnd[1], 1.0 )

      self.assertAlmostEqual( pin[0][0], mygrf.markx[0] )
      self.assertAlmostEqual( pin[0][1], mygrf.markx[1] )
      self.assertAlmostEqual( pin[0][2], mygrf.markx[2] )
      self.assertAlmostEqual( pin[1][0], mygrf.marky[0] )
      self.assertAlmostEqual( pin[1][1], mygrf.marky[1] )
      self.assertAlmostEqual( pin[1][2], mygrf.marky[2] )
      self.assertEqual( 1, mygrf.markt[0] )
      self.assertEqual( 3, mygrf.nmark )

      mygrf.Reset()
      plot.text( "Hello", [0.5,0.5], [0.0,1.0], "CC" )
      self.assertEqual( mygrf.textt, ['Hello'] )
      self.assertEqual( mygrf.textx, [0.5] )
      self.assertEqual( mygrf.texty, [0.5] )
      self.assertEqual( mygrf.textj, ['CC'] )
      self.assertEqual( mygrf.ntext, 1 )

   def test_MatrixMap(self):

      with self.assertRaises(ValueError):
         mm = starlink.Ast.MatrixMap( [ [ [1.0] ] ]  )

      mm = starlink.Ast.MatrixMap( [ -1.0, 2.0 ]  )
      self.assertIsInstance( mm, starlink.Ast.Object)
      self.assertIsInstance( mm, starlink.Ast.Mapping )
      self.assertIsInstance( mm, starlink.Ast.MatrixMap )
      self.assertTrue(  mm.isaobject() )
      self.assertTrue(  mm.isamapping() )
      self.assertTrue(  mm.isamatrixmap() )
      self.assertEqual( mm.Nin, 2 )
      self.assertEqual( mm.Nout, 2 )
      pin = numpy.array( [[1.,2.,3], [0.,1.,2]] )
      pout = mm.tran( pin, False )
      self.assertEqual( pout[0][0], -1.0 )
      self.assertEqual( pout[0][1], -2.0 )
      self.assertEqual( pout[0][2], -3.0 )
      self.assertEqual( pout[1][0], 0.0/2.0 )
      self.assertEqual( pout[1][1], 1.0/2.0 )
      self.assertEqual( pout[1][2], 2.0/2.0 )

      mm = starlink.Ast.MatrixMap( [[0.0,1.0], [2.0,3.0], [-1.0, -2.0]] )
      self.assertTrue( mm.TranForward )
      self.assertFalse( mm.TranInverse )
      self.assertEqual( mm.Nout, 3 )
      self.assertEqual( mm.Nin, 2 )
      pout = mm.tran( pin, True )
      self.assertEqual( pout[0][0], 0.0 )
      self.assertEqual( pout[0][1], 1.0 )
      self.assertEqual( pout[0][2], 2.0 )
      self.assertEqual( pout[1][0], 2.0 )
      self.assertEqual( pout[1][1], 7.0 )
      self.assertEqual( pout[1][2], 12.0 )
      self.assertEqual( pout[2][0], -1.0 )
      self.assertEqual( pout[2][1], -4.0 )
      self.assertEqual( pout[2][2], -7.0 )


   def test_PolyMap(self):
      pm = starlink.Ast.PolyMap( [[1.2,1.,2.,0.],[-0.5,1.,1.,1.],
                                  [1.0,2.,0.,1.]])
      self.assertIsInstance( pm, starlink.Ast.Object)
      self.assertIsInstance( pm, starlink.Ast.Mapping )
      self.assertIsInstance( pm, starlink.Ast.PolyMap )
      self.assertTrue(  pm.isaobject() )
      self.assertTrue(  pm.isamapping() )
      self.assertTrue(  pm.isapolymap() )
      self.assertEqual( pm.Nin, 2 )
      self.assertEqual( pm.Nout, 2 )
      pin = numpy.array( [[1.,2.,3], [0.,1.,2]] )
      pout = pm.tran( pin, True )
      for (xi,yi,xo,yo) in zip(pin[0],pin[1],pout[0],pout[1]):
         xn = 1.2*xi*xi - 0.5*yi*xi
         yn = yi
         self.assertAlmostEqual( xn, xo )
         self.assertAlmostEqual( yn, yo )

      pm = starlink.Ast.PolyMap( None, [[1.2,1.,2.,0.],[-0.5,1.,1.,1.],
                                        [1.0,2.,0.,1.]])
      self.assertEqual( pm.Nin, 2 )
      self.assertEqual( pm.Nout, 2 )
      pout = pm.tran( pin, False )
      for (xi,yi,xo,yo) in zip(pin[0],pin[1],pout[0],pout[1]):
         xn = 1.2*xi*xi - 0.5*yi*xi
         yn = yi
         self.assertAlmostEqual( xn, xo )
         self.assertAlmostEqual( yn, yo )

      pm = starlink.Ast.PolyMap( [[ 1., 1., 1., 0.],
                                  [ 1., 1., 0., 1.],
                                  [ 1., 2., 1., 0.],
                                  [-1., 2., 0., 1.]],
                                 [[ 0.5, 1., 1., 0.],
                                  [ 0.5, 1., 0., 1.],
                                  [ 0.5, 2., 1., 0.],
                                  [-0.5, 2., 0., 1.]] )

      self.assertEqual( pm.Nin, 2 )
      self.assertEqual( pm.Nout, 2 )
      pout = pm.tran( pin, True )
      pnew = pm.tran( pout, False )

      for (xi,yi,xn,yn) in zip(pin[0],pin[1],pnew[0],pnew[1]):
         self.assertAlmostEqual( xn, xi )
         self.assertAlmostEqual( yn, yi )

      self.assertEqual( pm.NiterInverse, 4 )
      self.assertEqual( pm.TolInverse, 1.0E-6 )

      new = pm.polytran( False, 1.0E-8, 0.01, 2, [-1.0, -1.0], [1.0, 1.0] )
      pout = new.tran( pin, True )
      pnew = new.tran( pout, False )
      for (xi,yi,xn,yn) in zip(pin[0],pin[1],pnew[0],pnew[1]):
         self.assertAlmostEqual( xn, xi )
         self.assertAlmostEqual( yn, yi )


   def test_MathMap(self):
      with self.assertRaises(TypeError):
         mathmap = starlink.Ast.MathMap( 2, 1, "r = sqrt( x * x + y * y )",
                                      (1.0, "y=r/sqrt(2)") )

      with self.assertRaises(TypeError):
         mathmap = starlink.Ast.MathMap( 2, 1, 1, ("x = r/sqrt(2)", "y=r/sqrt(2)") )

      mathmap = starlink.Ast.MathMap( 2, 1, "r = sqrt( x * x + y * y )",
                                      ("x = r", "y=r") )
      self.assertEqual( mathmap.Class, "MathMap" )
      self.assertIsInstance( mathmap, starlink.Ast.MathMap )
      self.assertEqual( mathmap.Nin, 2 )
      self.assertEqual( mathmap.Nout, 1 )

      pin = numpy.array( [[1.,2.,3], [0.,1.,2]] )
      pout = mathmap.tran( pin, True )
      for (x,y,r) in zip( pin[0], pin[1], pout[0] ):
         rn = math.sqrt( x*x + y*y )
         self.assertAlmostEqual( rn, r )

      pin2 = mathmap.tran( pout, False )
      for (r,x,y) in zip( pout[0], pin2[0], pin2[1] ):
         self.assertAlmostEqual( x, r )
         self.assertAlmostEqual( y, r )

   def test_Polygon(self):
      polygon = starlink.Ast.Polygon( starlink.Ast.Frame(2),
                                    [[0,1,0],[0,1,2]] )
      self.assertIsInstance( polygon, starlink.Ast.Polygon )
      self.assertIsInstance( polygon, starlink.Ast.Region )
      self.assertIsInstance( polygon, starlink.Ast.Frame )
      self.assertIsInstance( polygon.getregionframe(), starlink.Ast.Frame )
      overlap = polygon.overlap( polygon )
      self.assertEqual( overlap, 5 )
      testpolygon = starlink.Ast.Polygon( starlink.Ast.Frame(2),
                                          [[0,1,0],[1,2,3]] )
      overlap = polygon.overlap( testpolygon )
      self.assertEqual( overlap, 4 )
      array = numpy.array([[0,0,0,0],[0,1,1,0],[0,1,1,0],[0,0,0,0]])
      new = starlink.Ast.outline( 1, starlink.Ast.EQ, array, [-1,2], [2,5],
                                  0.0, 4, [0,3], True )
      pin = numpy.array( [[0.,0.,0], [1.9,2.1,4.5]] )
      pout = new.tran( pin, True )
      self.assertEqual( pout[0][0], starlink.Ast.BAD )
      self.assertEqual( pout[0][1], 0. )
      self.assertEqual( pout[0][2], starlink.Ast.BAD )
      self.assertEqual( pout[1][0], starlink.Ast.BAD )
      self.assertEqual( pout[1][1], 2.1 )
      self.assertEqual( pout[1][2], starlink.Ast.BAD )

      new2 = new.downsize( 0, 3 )
      self.assertTrue( new2.isapolygon() )

   def test_PointList(self):
      pointlist = starlink.Ast.PointList( starlink.Ast.Frame(2),
                                    [[0,1,0],[0,1,2]] )
      self.assertIsInstance( pointlist, starlink.Ast.PointList )
      self.assertIsInstance( pointlist, starlink.Ast.Region )
      self.assertIsInstance( pointlist, starlink.Ast.Frame )
      self.assertIsInstance( pointlist.getregionframe(), starlink.Ast.Frame )
      overlap = pointlist.overlap( pointlist )
      self.assertEqual( overlap, 5 )
      self.assertEqual( pointlist.ListSize, 3 )

   def test_Table(self):
      table = starlink.Ast.Table()
      self.assertIsInstance( table, starlink.Ast.Table )
      self.assertIsInstance( table, starlink.Ast.KeyMap )
      with self.assertRaises(starlink.Ast.BADKEY):
         table['Fred'] = 123
      dims = [5,2]
      table.addcolumn( 'Fred', starlink.Ast.DOUBLETYPE, dims )
      self.assertEqual( table.Ncolumn, 1 )
      self.assertEqual( table.columnname(1), 'FRED' )
      self.assertEqual( int(table.get( 'ColumnType(Fred)' )), starlink.Ast.DOUBLETYPE )
      dims = table.columnshape( 'Fred' )
      self.assertEqual( dims[0], 5 )
      self.assertEqual( dims[1], 2 )
      self.assertEqual( table.get( 'ColumnUnit(Fred)' ), '' )
      with self.assertRaises(starlink.Ast.BADTYP):
         table['Fred(2)'] = 123
      with self.assertRaises(starlink.Ast.BADTYP):
         table['Fred(2)'] = 123.0
      table['Fred(2)'] = numpy.linspace( 1.0, 10.0, 10.0 )
      self.assertEqual( table.columnlength('Fred'), 10 )
      self.assertEqual( table.columnndim('Fred'), 2 )
      self.assertEqual( table.columnunit('Fred'), '' )

   def test_FitsTable(self):
      table = starlink.Ast.FitsTable()
      self.assertIsInstance( table, starlink.Ast.FitsTable )
      self.assertIsInstance( table, starlink.Ast.Table )
      self.assertIsInstance( table, starlink.Ast.KeyMap )
      fc = starlink.Ast.FitsChan()
      fc.puttable( table, "Fred" )
      tables = fc.gettables()
      self.assertEqual(len(tables),1)
      keys = tables.keys()
      self.assertEqual( keys[0], 'Fred' )


if __name__ == "__main__":
#    starlink.Ast.watchmemory(10914)
    #unittest.main()
    suite = unittest.TestLoader().loadTestsFromTestCase(TestAst)
    unittest.TextTestRunner(verbosity=2).run(suite)
    starlink.Ast.activememory("AST memory block still active")



#!python3

"""
Generates the file exceptions.c which encapsulates Python/AST
exception handling.

The environment variable AST_SOURCE should be set to point to
the folder containing the source distribution for the AST
library.
"""

from __future__ import print_function

import os
import os.path

def make_exceptions( dirname=None ):

    if 'AST_SOURCE' not in os.environ:
        print("Please set AST_SOURCE environment variable to point to the AST source code directory")
        exit(1)

    # ensure that we have the error codes file
    errfile = os.path.join( os.environ['AST_SOURCE'], "ast_err.h")
    if not os.path.exists(errfile):
        print("Could not find the ast_err.h file in directory "+os.environ['AST_SOURCE'])
        exit(1)

    # Open an output C file
    cfilename = "exceptions.c"
    if dirname is not None:
        cfilename = os.path.join( dirname, cfilename )

    cfile = open( cfilename, "w" )

    # Need a C header
    print(r"""/*
*  Name:
*     exceptions.c

*  Purpose:
*     Raise a Python exception when AST reports an error.

*  Description:
*     For each AST error code defined in ast_err.h, this file creates
*     a singleton instance of a corresponding Python Exception class. It
*     also provides an implementation of the astPutErr function that AST
*     uses to report errors. This implementation raises the corresponding
*     Python exception whenever AST reports an error.

*  Notes:
*     - This file is generated automatically by the "make_exceptions.py"
*       script, and should not be edited.

*/

/* Prototypes for functions defined in this file. */
static int RegisterErrors( PyObject *m );

/* Declare a static variable to hold an instance of a base AST Exception
   from which all the others are derived. */
static PyObject *AstError_err;

/* For each AST error code, declare a static variable to hold an instance
   of the corresponding Python Exception. */""", file=cfile)

    # Now read the ast_err.h file and create extract all the error codes
    # Note that AST__3DFSET is not currently supported because a
    # variable can not start with a number
    errcodes = []
    for line in open( errfile,"r" ):
        words = line.split()
        if words and words[0] == "enum" and words[2][:5] == "AST__" and words[2][5:6].isalpha():
            errcodes.append( words[2][5:] )

    if not errcodes:
        print("Could not find any error codes. Aborting")
        cfile.close()
        os.path.unlink(cfilename)
        exit(1)

    for code in errcodes:
        print( "static PyObject *{0}_err;".format(code), file=cfile )

    print(r"""
/* Defines a function that creates a Python Exception object
   for each AST error code, and uses them to initialises the
   above static variables. It reurns 1 if successful, and zero
   if an error occurs. */

static int RegisterErrors( PyObject *m ){

   PyObject *dict = PyModule_GetDict(m);

/* First create an instance of a base AST exception class from which
   the other derive. */
   if( !( AstError_err = PyErr_NewException("Ast.AstError", NULL, NULL))) return 0;
   PyDict_SetItemString( dict, "AstError", AstError_err );

/* Now create an instance of each derived AST exception class. */""", file=cfile)

    for code in errcodes:
        print("   if( !({0}_err = PyErr_NewException(\"Ast.{0}\", AstError_err, NULL))) return 0;".format(code), file=cfile)
        print("   PyDict_SetItemString( dict, \"{0}\", {0}_err );".format(code), file=cfile)
        print(" ", file=cfile)

    print(r"""   return 1;
}


/* The AST library calls this function to deliver an error message
   in the form of a Python Exception. For each AST error code, raise
   the corresponding Python Exception. */
void astPutErr_( int status_value, const char *message ) {

/* Local Variables: */
   PyObject *ex;
   PyObject *ptype;
   PyObject *pvalue;
   PyObject *ptraceback;
   char *text;
   int lstat;
   int nc;

/* Clear the AST status value so that AST memory functions will function
   properly. */
   lstat = astStatus;
   astClearStatus;

/* If an AST exception has already occurred, append the new message to it. */
   ex = PyErr_Occurred();
   if( ex ){
      if( PyErr_GivenExceptionMatches( ex, AstError_err ) ) {

/* Get the existing Exception text */
         PyErr_Fetch( &ptype, &pvalue, &ptraceback );
         text = GetString( NULL, pvalue );
         if( text ) {

/* Ignore messages that give the C source file and line number since they are
   not useful for Python users. */
            if( strstr( text, "Ast.c" ) ) {
               text = astFree( text );
               nc = 0;
            } else {
               nc = strlen( text );
               text = astAppendString( text, &nc, "\n" );
            }

/* Append the new message and store the extended text with the Exception. */
            text = astAppendString( text, &nc, message );
            Py_DECREF( pvalue );
            pvalue = PyUnicode_FromString( text );
            PyErr_Restore( ptype, pvalue, ptraceback);
            text = astFree( text );
         }
      }

/* If an AST or non-AST exception has already occurred, return without action. */
      return;
   }

/* If no exception has already occurred, raise an appropriate AST exception now. */""", file=cfile)

    first = True
    for code in errcodes:
        if first:
            print("   if( status_value == AST__{0} ) {{".format(code), file=cfile)
            first = False
        else:
            print("   }} else if( status_value == AST__{0} ) {{".format(code), file=cfile)
        print("      PyErr_SetString( {0}_err, message );".format(code), file=cfile)

    print("""   } else {
      PyErr_SetString( AstError_err, message );
   }

/* restore the original AST status value. */
   astSetStatus( lstat );
}
""", file=cfile)
    cfile.close()

if __name__ == "__main__":
    make_exceptions()
